// Copyright (c) 1999-2015  Pavel Rychly, Milos Jakubicek

#ifndef FINLIB_REGEXPLEX_HH
#define FINLIB_REGEXPLEX_HH

#include <finlib/fsop.hh>
#include <finlib/regpref.hh>
#include <finlib/generator.hh>
#include <finlib/config.hh>

#if defined FINLIB_DARWIN
#include <libiberty.h>
#endif

template <class Revidx, class Lexicon>
FastStream *regexp2poss (Revidx &rev, Lexicon &lex, const char *pattern, 
                         const char *locale = NULL, const char *encoding = NULL, 
                         bool ignorecase = false, FastStream* filter_fs = NULL)
{
    regexp_pattern pat (pattern, locale, encoding, ignorecase);

    if (pat.any())
        return new SequenceStream (0, rev.maxpos() - 1, rev.maxpos());

    if (pat.no_meta_chars()) {
        // pat.prefix has been de-escaped
        int id = lex.str2id (pat.prefix());
        if (id < 0)
            return new EmptyStream();
        else
            return rev.id2poss (id);
    }

    vector<string> *parts = pat.disjoint_parts();
    if (parts->size() && !ignorecase) {
        // pattern has form (word1|word2|word3|...)
        vector<FastStream*> *fsv = new std::vector<FastStream*>;
        for (size_t i = 0; i < parts->size(); i++) {
            int id = lex.str2id ((*parts)[i].c_str());
            if (id >= 0)
                fsv->push_back (rev.id2poss (id));
        }
        return QOrVNode::create (fsv);
    }

    if (pat.compile())
        return new EmptyStream();

    if (!filter_fs) {
        Generator<int> *ids = lex.pref2ids (pat.prefix());
        if (ids->end()) {
            delete ids;
            return new EmptyStream();
        }
        filter_fs = new Gen2Fast<int> (ids);
    }

    std::vector<FastStream*> *fsv = new std::vector<FastStream*>;
    fsv->reserve(32);
    Position id, finval = filter_fs->final();
    while ((id = filter_fs->next()) < finval) {
        if (pat.match (lex.id2str (id)))
            fsv->push_back(rev.id2poss (id));
    }
    delete filter_fs;
    return QOrVNode::create (fsv);
}

template <class Revidx, class Lexicon>
FastStream *compare2poss (Revidx &rev, Lexicon &lex, const char *pattern,
                         int order, bool ignorecase = false)
{
    std::vector<FastStream*> *fsv = new std::vector<FastStream*>;
    fsv->reserve(32);
    for (int id = 0; id < lex.size(); id++) {
        const char *str = lex.id2str (id);
        int cmp = strverscmp(str, pattern);
        if ((order < 0 && cmp <= 0) || (order > 0 && cmp >= 0)) {
            FastStream *s = rev.id2poss (id);
            fsv->push_back(s);
        }
    }
    return QOrVNode::create (fsv);
}

template <class Lexicon>
class regexp2idsStream : public Generator<int>
{
protected:
    Lexicon &lex;
    Generator<int> *lexit;
    regexp_pattern *pat;
    int currid;
    bool finished;
    bool negative;
    void locate () {
        finished = true;
        while (!lexit->end()) {
            currid = lexit->next();
            if (pat->match (lex.id2str (currid)) == !negative) {
                finished = false;
                break;
            }
        }
    }
public:
    regexp2idsStream (Lexicon &l, Generator<int> *li, regexp_pattern *p, bool negative = false)
        : lex (l), lexit (li), pat (p), finished (false), negative(negative) {
        locate();
    }
    virtual int next() {
        int ret = currid;
        locate(); 
        return ret;
    }
    virtual bool end() {return finished;}
    virtual int size() {return finished ? 0 : lexit->size() + 1;}
    virtual ~regexp2idsStream() {delete lexit; delete pat;}
};


template <class Lexicon>
Generator<int> *regexp2ids (Lexicon &lex, const char *pattern,
        const char *locale = NULL, const char *encoding = NULL,
        bool ignorecase = false, const char *filter_pattern = NULL,
        FastStream *filter_fs = NULL)
{
    regexp_pattern *pat = new regexp_pattern (pattern, locale, encoding, ignorecase);
    Generator<int> *ids;
    vector<string> *parts = pat->disjoint_parts();

    if (pat->any()) {
        ids = new SequenceGenerator<int> (0, lex.size() -1);
    } else if (pat->no_meta_chars() && !ignorecase) {
        // pattern is no regexp, pat.prefix() has been de-escaped
        int id = lex.str2id (pat->prefix());
        delete pat;
        if (id < 0)
            return new EmptyGenerator<int>();
        else
            ids = new SequenceGenerator<int> (id, id);
    } else if (parts->size() && !ignorecase) {
        // pattern has form (word1|word2|word3|...)
        int *str_ids = new int[parts->size()];
        size_t i;
        for (i = 0; i < parts->size(); i++) {
            int id = lex.str2id ((*parts)[i].c_str());
            if (id >= 0)
                str_ids[i] = id;
        }
        switch (i) {
        case 0:
            delete[] str_ids;
            ids = new EmptyGenerator<int>();
        case 1: {
            delete[] str_ids;
            ids = new SequenceGenerator<int> (str_ids[0], str_ids[0]);
        }
        default:
            ids = new ArrayGenerator<int> (str_ids, i);
        }
    } else {
        if (pat->compile()) {
            delete pat;
            return new EmptyGenerator<int>();
        }
        if (ignorecase)
            ids = new SequenceGenerator<int> (0, lex.size() -1);
        else {// try non-regexp prefix
            if (filter_fs) {
                if (filter_fs->peek() < filter_fs->final())
                    ids = new Fast2Gen<int> (filter_fs);
                else {
                    delete pat;
                    return new EmptyGenerator<int>();
                }
            }
            else {
                ids = lex.pref2ids (pat->prefix());
                if (ids->end()) {
                    delete pat;
                    delete ids;
                    return new EmptyGenerator<int>();
                }
            }
        }
        ids = new regexp2idsStream<Lexicon> (lex, ids, pat);
    }

    if (filter_pattern) {
        regexp_pattern *filter = new regexp_pattern (filter_pattern, locale, encoding, ignorecase);
        if (filter->compile()) {
            delete filter;
            return new EmptyGenerator<int>();
        }
        return new regexp2idsStream<Lexicon> (lex, ids, filter, true);
    } else {
        return ids;
    }
}

#endif

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:

